#include <stdio.h>
#include "hoc6.h"
#include "y.tab.h"

#define NSTACK 256
#define NPROG  2000
#define NFRAME 100

#define TRACE  0             /* printing the names of the executing function */
#define SYM    -1            /* to distinguish values and names in dump_stack */
#define VAL    -2

static Datum  stack[NSTACK];        /* "THE" stack */
static Datum *stackp;               /* next open entry */
static int    tstack[NSTACK];       /* type of union element for dump_stack */

int     iprog, istack; /* for dump_stack, dump_prog */

Inst    prog[NPROG];                /* Machine stack */
Inst    *progp;                     /* next open entry on stack  */
Inst    *pc;                        /* next instruction on execution */
Inst    *progbase = prog;           /* start of current subprogram */
int     returning;                  /* ==1 if return stmt is seen */

typedef struct Frame                /* proc/func call stack frame */
{   Symbol  *sp;                    /* symbol table entry */
    Inst    *retpc;                 /* where to resume after return */
    Datum   *argn;                  /* n-th argument on stack */
    int     nargs;                  /* number of arguments */
} Frame;

Frame   frame[NFRAME];
Frame   *fp;

initcode()
{
    if (TRACE) fprintf(stderr,"calling initcode\n");
    stackp = stack;
    progp  = progbase;
    fp = frame;
    returning = 0;
    iprog = istack = 0;
    printf("hoc> ");
}

push(Datum d)
{
    if (TRACE) fprintf(stderr,"calling push\n");
    if (stackp >= &stack[NSTACK])
        execerror("stack overflow",(char *)0);
    *stackp++ = d;
    istack++;
}

Datum pop()
{
    if (TRACE) fprintf(stderr,"calling pop\n");
    if(stackp <= stack) 
        execerror("stack underflow",(char *)0);
    istack--;
    return *--stackp;
}
#if 0
void dump_machine()
{
    int     i;
    
    for(i=iprog-1; i>=0; i--)
    {
    }
}

void dump_stack()
{
    int     i;

    printf("i:\tval/name:\n");
    for ( i=istack-1; i>=0; i--)
    {   
        if(tstack[i] == SYM)
            printf("%4d\t%s\n", i,stack[i].sym->name);
        else
            printf("%4d\t%.4g\n", i,stack[i].val);
    }
}
#endif

constpush()
{
    Datum   d;

    if (TRACE) fprintf(stderr,"calling constpush\n");
    d.val = ((Symbol *)*pc++)->u.val;
    push(d);
    tstack[istack-1] = VAL;
}

varpush()
{
    Datum   d;
    if (TRACE) fprintf(stderr,"calling varpush\n");
    d.sym = (Symbol *)(*pc++);
    push(d);
    tstack[istack-1] = SYM;
}

whilecode()
{
    Datum d;
    Inst  *savepc = pc;                /* loop body */

    if (TRACE) fprintf(stderr,"calling whilecode\n");
    execute(savepc+2);                 /* eval condition */
    d = pop();
    while(d.val)                       /* cond true */
    {   execute(*((Inst **)(savepc))); /* while-block */
        if(returning)
            break;
        execute(savepc+2);             /* eval condition */
        d = pop();
    }
    if(!returning)
        pc = *((Inst **)(savepc+1));   /* next stmt after while */
}

ifcode()
{
    Datum d;
    Inst  *savepc = pc;                /* loop body */

    if (TRACE) fprintf(stderr,"calling ifcode\n");
    execute(savepc+3);                 /* eval condition */
    d = pop();
    if (d.val)                         /* cond true */
        execute(*((Inst **)(savepc))); /* if-block */
    else if (*((Inst **)(savepc+1)))   /* else-block ??? */
        execute(*((Inst **)(savepc+1))); /* else-block */
    if(!returning)
        pc = *((Inst **)(savepc+2));   /* next stmt after if */
}

define(Symbol *sp)                     /* put func/proc in symbol table */
{
    if (TRACE) fprintf(stderr,"calling define\n");
    sp->u.defn = (Inst)progbase;       /* start of code */
    progbase = progp;                  /* next code starts here */
}

call()
{
    Symbol *sp = (Symbol *)pc[0];      /* symbol table entry for function */
    if (TRACE) fprintf(stderr,"calling call\n");
    if(fp++ >= &frame[NFRAME-1])
        execerror(sp->name,"call nested too deeply");
    fp->sp    = sp;
    fp->nargs = (int)pc[1];
    fp->retpc = pc + 2;
    fp->argn  = stackp - 1;
    execute(sp->u.defn);
    returning = 0;
}

ret()                                  /* common return from func/proc */
{
    int i;
    if (TRACE) fprintf(stderr,"calling ret\n");
    for(i=0; i<fp->nargs; i++)
        pop();                         /* pop arguments */
    pc = (Inst *)fp->retpc;
    --fp;
    returning = 1;    
}

funcret()
{
    Datum d;
    if (TRACE) fprintf(stderr,"calling funcret\n");
    if(fp->sp->type == PROCEDURE)
        execerror(fp->sp->name, "(proc) returns value");
    d = pop();                         /* preserve function return value */
    ret();
    push(d);
    tstack[istack-1] = VAL;
}

procret()
{
    Datum d;
    if (TRACE) fprintf(stderr,"calling procret\n");
    if(fp->sp->type == FUNCTION)
        execerror(fp->sp->name, "(func) returns no value");
    ret();
}

double *getarg()
{
    int nargs = (int)*pc++;
    if (TRACE) fprintf(stderr,"calling getarg\n");
    if(nargs > fp->nargs)
        execerror(fp->sp->name,"not enough arguments");
    return &fp->argn[nargs - fp->nargs].val;
}

arg()
{
    Datum d;
    if (TRACE) fprintf(stderr,"calling arg\n");
    d.val = *getarg();
    push(d);
    tstack[istack-1] = VAL;
}

argassign()
{
    Datum d;
    if (TRACE) fprintf(stderr,"calling argassign\n");
    d = pop();
    push(d); /* leave value on stack */
    tstack[istack-1] = VAL;
    *getarg() = d.val;
}

add()
{
    Datum   d1,d2;
    if (TRACE) fprintf(stderr,"calling add\n");
    d2 = pop();
    d1 = pop();
    d1.val += d2.val;
    push(d1);
    tstack[istack-1] = VAL;
}

sub()
{
    Datum   d1,d2;
    if (TRACE) fprintf(stderr,"calling sub\n");
    d2 = pop();
    d1 = pop();
    d1.val -= d2.val;
    push(d1);
    tstack[istack-1] = VAL;
}

mul()
{
    Datum   d1,d2;
    if (TRACE) fprintf(stderr,"calling mul\n");
    d2 = pop();
    d1 = pop();
    d1.val *= d2.val;
    push(d1);
    tstack[istack-1] = VAL;
}

mod()
{
    Datum   d1,d2;
    if (TRACE) fprintf(stderr,"calling mul\n");
    d2 = pop();
    d1 = pop();
    d1.val = (double)((int)d1.val % (int)d2.val);
    push(d1);
    tstack[istack-1] = VAL;
}

ddiv()
{
    Datum   d1,d2;
    if (TRACE) fprintf(stderr,"calling ddiv\n");
    d2 = pop();
    if (d2.val==0.0)
        execerror("division by zero",(char *)0);
    d1 = pop();
    d1.val /= d2.val;
    push(d1);
    tstack[istack-1] = VAL;
}

power()
{   double Pow(double x, double y);

    Datum   d1,d2;
    if (TRACE) fprintf(stderr,"calling power\n");
    d2 = pop();
    d1 = pop();
    d1.val = Pow(d1.val,d2.val);
    push(d1);
    tstack[istack-1] = VAL;
}

negate()
{
    Datum   d;
    if (TRACE) fprintf(stderr,"calling negate\n");
    d = pop();
    d.val = -d.val;
    push(d);
    tstack[istack-1] = VAL;
}

eval()
{
    Datum   d;
    if (TRACE) fprintf(stderr,"calling eval\n");
    d = pop();
    if (d.sym->type != VAR && d.sym->type != UNDEF)
        execerror("attempt to evaluate non-variable", d.sym->name);
    if (d.sym->type == UNDEF)
        execerror("undefined variable", d.sym->name);
    d.val = d.sym->u.val;
    push(d);
    tstack[istack-1] = VAL;
}

assign()
{
    Datum   d1,d2;
    if (TRACE) fprintf(stderr,"calling assign\n");
    d1 = pop();
    d2 = pop();
    if (d1.sym->type != VAR && d1.sym->type != UNDEF)
        execerror("assignment to non-variable", d1.sym->name);
    d1.sym->u.val = d2.val;
    d1.sym->type  = VAR;
    push(d2);
    tstack[istack-1] = VAL;
}

print()
{
    Datum   d;
    if (TRACE) fprintf(stderr,"calling print\n");
    d = pop();
    /*printf("-->\t%.8g\n",d.val);*/
    printf("\t%.8g\n",d.val);
}

prexpr()
{
    Datum   d;
    if (TRACE) fprintf(stderr,"calling prexpr\n");
    d = pop();
    printf("%.8g",d.val);
}

prstr()
{
    if (TRACE) fprintf(stderr,"calling prstr\n");
    printf("%s",(char *)*pc++);
}

varread()                         /* read value into variable */
{
    Datum d;
    extern FILE *yyin;
    Symbol *var = (Symbol *)*pc++;
    if (TRACE) fprintf(stderr,"calling varread\n");
Again:
    switch(fscanf(yyin,"%lf",&var->u.val))
    { 
      case EOF:
        if (yywrap()==0)
            goto Again;
        d.val = var->u.val = 0.0;
        break;
      case 0:
        execerror("non-number read into",var->name);
        break;
      default:
        d.val = 1.0;
        break;
    }
    var->type = VAR;
    push(d);
    tstack[istack-1] = VAL;
}


bltin()
{
    Datum   d;
    if (TRACE) fprintf(stderr,"calling bltin\n");
    d = pop();
    d.val = (*(double (*)())(*pc++))(d.val);
    push(d);
    tstack[istack-1] = VAL;
}

le()
{
    Datum d1,d2;
    if (TRACE) fprintf(stderr,"calling le\n");
    d2 = pop();
    d1 = pop();
    d1.val = (double)(d1.val <= d2.val);
    push(d1);
    tstack[istack-1] = VAL;
}

lt()
{
    Datum d1,d2;
    if (TRACE) fprintf(stderr,"calling lt\n");
    d2 = pop();
    d1 = pop();
    d1.val = (double)(d1.val < d2.val);
    push(d1);
    tstack[istack-1] = VAL;
}

ge()
{
    Datum d1,d2;
    if (TRACE) fprintf(stderr,"calling ge\n");
    d2 = pop();
    d1 = pop();
    d1.val = (double)(d1.val >= d2.val);
    push(d1);
    tstack[istack-1] = VAL;
}

gt()
{
    Datum d1,d2;
    if (TRACE) fprintf(stderr,"calling gt\n");
    d2 = pop();
    d1 = pop();
    d1.val = (double)(d1.val > d2.val);
    push(d1);
    tstack[istack-1] = VAL;
}

eq()
{
    Datum d1,d2;
    if (TRACE) fprintf(stderr,"calling eq\n");
    d2 = pop();
    d1 = pop();
    d1.val = (double)(d1.val == d2.val);
    push(d1);
    tstack[istack-1] = VAL;
}

ne()
{
    Datum d1,d2;
    if (TRACE) fprintf(stderr,"calling ne\n");
    d2 = pop();
    d1 = pop();
    d1.val = (double)(d1.val != d2.val);
    push(d1);
    tstack[istack-1] = VAL;
}

and()
{
    Datum d1,d2;
    if (TRACE) fprintf(stderr,"calling and\n");
    d2 = pop();
    d1 = pop();
    d1.val = (double)(d1.val!=0.0 && d2.val!=0.0);
    push(d1);
    tstack[istack-1] = VAL;
}

or()
{
    Datum d1,d2;
    if (TRACE) fprintf(stderr,"calling or\n");
    d2 = pop();
    d1 = pop();
    d1.val = (double)(d1.val!=0.0 || d2.val!=0.0);
    push(d1);
    tstack[istack-1] = VAL;
}

not()
{
    Datum d1;
    if (TRACE) fprintf(stderr,"calling not\n");
    d1 = pop();
    d1.val = (double)(d1.val!=0.0);
    push(d1);
    tstack[istack-1] = VAL;
}

Inst *code(Inst f)
{
    Inst *oprogp = progp;
    /*
    if (TRACE) fprintf(stderr,"calling code\n");
    fprintf(stderr,"code(%d)\n",*f);
    */
    if(progp>= &prog[NPROG])
        execerror("program too big",(char *)0);
    *progp++ = f;
    iprog++;
    return oprogp;
}

execute(Inst *p)
{    
#if 0
    void dump_stack();
#endif
    if (TRACE) fprintf(stderr,"calling execute\n");
    for(pc=p; *pc!=STOP && !returning; )
    {   (*(*pc++))();
#if 0
        dump_stack();
#endif
    }
}

